#include <complex.h>
#include <math.h>
#include <omp.h>
#include <stdbool.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

// CLASS = C
const int NX = 512;
const int NY = 512;
const int NZ = 512;
const int MAXDIM = 512;
const int NITER_DEFAULT = 20;

// total number of grid points with padding
const int NXP = NX + 1;
const int NTOTALP = NXP * NY * NZ;
const double NTOTAL_F = (double)NX * NY * NZ;

// If processor array is 1x1 -> 0D grid decomposition

// Cache blocking params. These values are good for most RISC processors.
// FFT parameters:
//   fftblock controls how many ffts are done at a time.
//   The default is appropriate for most cache-based machines
//   On vector machines, the FFT can be vectorized with vector
//   length equal to the block size, so the block size should
//   be as large as possible. This is the size of the smallest
//   dimension of the problem: 128 for class A, 256 for class B and
//   512 for class C.

const int FFTBLOCK_DEFAULT = 32;
const int FFTBLOCKPAD_DEFAULT = FFTBLOCK_DEFAULT + 2;

// other stuff
const int SEED = 314159265;
const double ALPHA = 1E-6;

struct FT {
	bool debug;

	int niter;
	int fftblock;
	int fftblockpad;

	double complex *u;
	double complex *u0;
	double complex *u1;
	double *twiddle;

	int d1;
	int d2;
	int d3;
};

// 2D-Index (i, j) with dimensions (di, ???)
static inline int ix2d(int i, int di, int j)
{
	return (j * di) + i;
}

// 3D-Index (i, j, k) with dimensions (di, dj, ???)
static inline int ix3d(int i, int di, int j, int dj, int k)
{
	return (k * dj * di) + ix2d(i, di, j);
}

// 3D-Index (i, j, k) with dimensions (ft.d1 + 1, ft.d2, ft.d3)
static inline int ixFT(struct FT ft, int i, int j, int k)
{
	return ix3d(i, ft.d1 + 1, j, ft.d2, k);
}

// 3D-Index (j, i, omp_get_thread_num()) with dimensions (ft.fftblockpad, di, omp_get_max_threads())
static inline int ixFTBlock(const struct FT ft, const int i, const int di, const int j)
{
	return ix3d(j, ft.fftblockpad, i, di, omp_get_thread_num());
}

static int ilog2(const int n)
{
	if (n == 1)
		return 0;

	int lg = 1;
	int nn = 2;

	while (nn < n) {
		nn *= 2;
		lg += 1;
	}

	return lg;
}

static void alloc_space(struct FT *ft)
{
	ft->u = malloc(NXP * sizeof(*ft->u));
	ft->u0 = malloc(NTOTALP * sizeof(*ft->u0));
	ft->u1 = malloc(NTOTALP * sizeof(*ft->u1));
	ft->twiddle = malloc(NTOTALP * sizeof(*ft->twiddle));

	if (ft->u0 == NULL || ft->u1 == NULL || ft->twiddle == NULL) {
		printf("Error encountered in allocating space\n");
		exit(EXIT_FAILURE);
	}
}

static struct FT setup(void)
{
	struct FT ft = {0};

	ft.debug = false;
	ft.niter = NITER_DEFAULT;

	printf("\n");
	printf("\n");
	printf(" NAS Parallel Benchmarks (NPB3.4-OMP) - FT Benchmark\n");
	printf("\n");
	printf(" Size                        : %dx%dx%d\n", NX, NY, NZ);
	printf(" Iterations                  : %d\n", ft.niter);
	printf(" Number of available threads : %d\n", omp_get_max_threads());
	printf("\n");

	ft.d1 = NX;
	ft.d2 = NY;
	ft.d3 = NZ;

	// ---------------------------------------------------------------------
	// Set up info for blocking of ffts and transposes.  This improves
	// performance on cache-based systems. Blocking involves
	// working on a chunk of the problem at a time, taking chunks
	// along the first, second, or third dimension.
	//
	// - In cffts1 blocking is on 2nd dimension (with fft on 1st dim)
	// - In cffts2/3 blocking is on 1st dimension (with fft on 2nd and 3rd dims)
	//
	// Since 1st dim is always in processor, we'll assume it's long enough
	// (default blocking factor is 16 so min size for 1st dim is 16)
	// The only case we have to worry about is cffts1 in a 2d decomposition.
	// so the blocking factor should not be larger than the 2nd dimension.
	// ---------------------------------------------------------------------

	ft.fftblock = FFTBLOCK_DEFAULT;
	ft.fftblockpad = FFTBLOCKPAD_DEFAULT;

	if (ft.fftblock != FFTBLOCK_DEFAULT)
		ft.fftblockpad = ft.fftblock + 3;

	alloc_space(&ft);

	return ft;
}

static void init_ui(const struct FT ft)
{
	// clang-format off
	#pragma omp parallel for collapse(2)
	// clang-format on
	for (int k = 0; k < ft.d3; k++) {
		for (int j = 0; j < ft.d2; j++) {
			for (int i = 0; i < ft.d1 + 1; i++) {
				// clang-format off
				#define index ixFT(ft, i, j, k)
				// clang-format on

				ft.u0[index] = 0;
				ft.u1[index] = 0;
				ft.twiddle[index] = 0;

				// clang-format off
				#undef index
				// clang-format on
			}
		}
	}
}

// ---------------------------------------------------------------------
//  compute function from local (i,j,k) to ibar^2+jbar^2+kbar^2
//  for time evolution exponent.
// ---------------------------------------------------------------------
static void compute_indexmap(const struct FT ft)
{
	const double AP = -4 * ALPHA * M_PI * M_PI;

	// ---------------------------------------------------------------------
	// basically we want to convert the fortran indices
	//   1 2 3 4 5 6 7 8
	// to
	//   0 1 2 3 -4 -3 -2 -1
	// The following magic formula does the trick:
	// mod(i-1+n/2, n) - n/2
	// ---------------------------------------------------------------------

	// clang-format off
	#pragma omp parallel for collapse(2)
	// clang-format on
	for (int k = 0; k < ft.d3; k++) {
		for (int j = 0; j < ft.d2; j++) {
			const int kk = (k + ft.d3 / 2) % ft.d3 - ft.d3 / 2;
			const int kk2 = kk * kk;

			const int jj = (j + ft.d2 / 2) % ft.d2 - ft.d2 / 2;
			const int kj2 = jj * jj + kk2;

			for (int i = 0; i < ft.d1; i++) {
				// clang-format off
				#define index ixFT(ft, i, j, k)
				// clang-format on

				const int ii = (i + ft.d1 / 2) % ft.d1 - ft.d1 / 2;
				ft.twiddle[index] = exp(AP * (ii * ii + kj2));

				// clang-format off
				#undef index
				// clang-format on
			}
		}
	}
}

static void compute_initial_conditions(const struct FT ft)
{
	srand48(SEED);

	for (int k = 0; k < ft.d3; k++) {
		for (int j = 0; j < ft.d2; j++) {
			for (int i = 0; i < ft.d1 + 1; i++) {
				// clang-format off
				#define index ixFT(ft, i, j, k)
				// clang-format on

				ft.u1[index] = drand48();

				// clang-format off
				#undef index
				// clang-format on
			}
		}
	}
}

static void evolve(struct FT ft)
{
	// ---------------------------------------------------------------------
	//  evolve u0 -> u1 (t time steps) in fourier space
	// ---------------------------------------------------------------------

	// clang-format off
	#pragma omp parallel for collapse(2)
	// clang-format on
	for (int k = 0; k < ft.d3; k++) {
		for (int j = 0; j < ft.d2; j++) {
			for (int i = 0; i < ft.d1; i++) {
				// clang-format off
				#define index ixFT(ft, i, j, k)
				// clang-format on

				ft.u0[index] *= ft.twiddle[index];
				ft.u1[index] = ft.u0[index];

				// clang-format off
				#undef index
				// clang-format on
			}
		}
	}
}

// ---------------------------------------------------------------------
//  compute the roots-of-unity array that will be used for subsequent FFTs.
// ---------------------------------------------------------------------
static void fft_init(const struct FT ft)
{
	// ---------------------------------------------------------------------
	//   Initialize the U array with sines and cosines in a manner that permits
	//   stride one access at each FFT iteration.
	// ---------------------------------------------------------------------

	const int m = ilog2(ft.d1 - 1);
	ft.u[0] = m;

	int ku = 1;
	int ln = 1;

	for (int j = 0; j < m; j++) {
		const double t = M_PI / ln;

		for (int i = 0; i < ln; i++) {
			const double ti = i * t;

			ft.u[i + ku] = cos(ti) + I * sin(ti);
		}

		ku += ln;
		ln *= 2;
	}
}

static void fftz2(const struct FT ft,
                  const int is,
                  const int l,
                  const int m,
                  const int n,
                  double complex *x,
                  double complex *y)
{
	// ---------------------------------------------------------------------
	//    Performs the L-th iteration of the second variant of the Stockham FFT.
	// --------------------------------------------------------------------

	const int n1 = n / 2;
	const int lk = 1 << (l - 1);
	const int li = 1 << (m - l);
	const int lj = 2 * lk;

	for (int i = 0; i < li; i++) {
		const int i11 = i * lk;
		const int i12 = i11 + n1;
		const int i21 = i * lj;
		const int i22 = i21 + lk;

		double complex u1;

		if (is >= 1)
			u1 = ft.u[li + i];
		else
			u1 = conj(ft.u[li + i]);

		// ---------------------------------------------------------------------
		//    This loop is vectorizable.
		// ---------------------------------------------------------------------
		for (int k = 0; k < lk; k++) {
			for (int j = 0; j < ft.fftblock; j++) {
				// clang-format off
				#define index11 ixFTBlock(ft, i11 + k, n, j)
				#define index12 ixFTBlock(ft, i12 + k, n, j)
				// clang-format on

				const double complex x11 = x[index11];
				const double complex x21 = x[index12];

				// clang-format off
				#undef index21
				#undef index22
				// clang-format on

				// clang-format off
				#define index21 ixFTBlock(ft, i21 + k, n, j)
				#define index22 ixFTBlock(ft, i22 + k, n, j)
				// clang-format on

				y[index21] = x11 + x21;
				y[index22] = u1 * (x11 - x21);

				// clang-format off
				#undef index11
				#undef index12
				// clang-format on
			}
		}
	}
}

static void cfftz(const struct FT ft,
                  const int is,
                  const int m,
                  const int n,
                  double complex *x,
                  double complex *y)
{
	// ---------------------------------------------------------------------
	//    Computes NY N-point complex-to-complex FFTs of X using an algorithm due
	//    to Swarztrauber.  X is both the input and the output array, while Y is a
	//    scratch array.  It is assumed that N = 2^M.  Before calling CFFTZ to
	//    perform FFTs, the array U must be initialized by calling CFFTZ with IS
	//    set to 0 and M set to MX, where MX is the maximum value of M for any
	//    subsequent call.
	// ---------------------------------------------------------------------

	// ---------------------------------------------------------------------
	//    Check if input parameters are invalid.
	// ---------------------------------------------------------------------

	const int mx = ft.u[0];

	if ((is != 1 && is != -1) || m < 1 || m > mx) {
		// clang-format off
		printf("CFFTZ: Either U has not been initialized, or else one "
		       "of the input parameters is invalid %d %d %d\n", is, m, mx);
		// clang-format on
		exit(EXIT_FAILURE);
	}

	// ---------------------------------------------------------------------
	//    Perform one variant of the Stockham FFT.
	// ---------------------------------------------------------------------

	for (int l = 1; l <= m; l += 2) {
		fftz2(ft, is, l, m, n, x, y);

		if (l == m) {
			// ---------------------------------------------------------------------
			//   Copy Y to X.
			// ---------------------------------------------------------------------

			for (int j = 0; j < n; j++) {
				for (int i = 0; i < ft.fftblock; i++) {
					// clang-format off
					#define index ixFTBlock(ft, j, n, i)
					// clang-format on

					x[index] = y[index];

					// clang-format off
					#undef index
					// clang-format on
				}
			}

			continue;
		}

		fftz2(ft, is, l + 1, m, n, y, x);
	}
}

static void cffts1(const struct FT ft,
                   const int is,
                   const double complex *x,
                   double complex *xout,
                   double complex *y1,
                   double complex *y2)
{
	const int logd1 = ilog2(ft.d1);

	// clang-format off
	#pragma omp parallel for collapse(2)
	// clang-format on
	for (int k = 0; k < ft.d3; k++) {
		for (int jn = 0; jn < ft.d2 / ft.fftblock; jn++) {
			const int jj = jn * ft.fftblock;

			for (int j = 0; j < ft.fftblock; j++) {
				for (int i = 0; i < ft.d1; i++) {
					// clang-format off
					#define indexY ixFTBlock(ft, i, ft.d1, j)
					#define indexX ixFT(ft, i, j + jj, k)
					// clang-format on

					y1[indexY] = x[indexX];

					// clang-format off
					#undef indexY
					#undef indexX
					// clang-format on
				}
			}

			cfftz(ft, is, logd1, ft.d1, y1, y2);

			for (int j = 0; j < ft.fftblock; j++) {
				for (int i = 0; i < ft.d1; i++) {
					// clang-format off
					#define indexY ixFTBlock(ft, i, ft.d1, j)
					#define indexX ixFT(ft, i, j + jj, k)
					// clang-format on

					xout[indexX] = y1[indexY];

					// clang-format off
					#undef indexY
					#undef indexX
					// clang-format on
				}
			}
		}
	}
}

static void cffts2(const struct FT ft,
                   const int is,
                   const double complex *x,
                   double complex *xout,
                   double complex *y1,
                   double complex *y2)
{
	const int logd2 = ilog2(ft.d2);

	// clang-format off
	#pragma omp parallel for collapse(2)
	// clang-format on
	for (int k = 0; k < ft.d3; k++) {
		for (int in = 0; in < ft.d1 / ft.fftblock; in++) {
			const int ii = in * ft.fftblock;

			for (int j = 0; j < ft.d2; j++) {
				for (int i = 0; i < ft.fftblock; i++) {
					// clang-format off
					#define indexY ixFTBlock(ft, j, ft.d2, i)
					#define indexX ixFT(ft, i + ii, j, k)
					// clang-format on

					y1[indexY] = x[indexX];

					// clang-format off
					#undef indexY
					#undef indexX
					// clang-format on
				}
			}

			cfftz(ft, is, logd2, ft.d2, y1, y2);

			for (int j = 0; j < ft.d2; j++) {
				for (int i = 0; i < ft.fftblock; i++) {
					// clang-format off
					#define indexY ixFTBlock(ft, j, ft.d2, i)
					#define indexX ixFT(ft, i + ii, j, k)
					// clang-format on

					xout[indexX] = y1[indexY];

					// clang-format off
					#undef indexY
					#undef indexX
					// clang-format on
				}
			}
		}
	}
}

static void cffts3(const struct FT ft,
                   const int is,
                   const double complex *x,
                   double complex *xout,
                   double complex *y1,
                   double complex *y2)
{
	const int logd3 = ilog2(ft.d3);

	// clang-format off
	#pragma omp parallel for collapse(2)
	// clang-format on
	for (int j = 0; j < ft.d2; j++) {
		for (int in = 0; in < ft.d1 / ft.fftblock; in++) {
			const int ii = in * ft.fftblock;

			for (int k = 0; k < ft.d3; k++) {
				for (int i = 0; i < ft.fftblock; i++) {
					// clang-format off
					#define indexY ixFTBlock(ft, k, ft.d3, i)
					#define indexX ixFT(ft, i + ii, j, k)
					// clang-format on

					y1[indexY] = x[indexX];

					// clang-format off
					#undef indexY
					#undef indexX
					// clang-format on
				}
			}

			cfftz(ft, is, logd3, ft.d3, y1, y2);

			for (int k = 0; k < ft.d3; k++) {
				for (int i = 0; i < ft.fftblock; i++) {
					// clang-format off
					#define indexY ixFTBlock(ft, k, ft.d3, i)
					#define indexX ixFT(ft, i + ii, j, k)
					// clang-format on

					xout[indexX] = y1[indexY];

					// clang-format off
					#undef indexY
					#undef indexX
					// clang-format on
				}
			}
		}
	}
}

void fft(const struct FT ft, const int dir, double complex *x1, double complex *x2)
{
	double complex y1[ft.fftblockpad * MAXDIM * omp_get_max_threads()];
	double complex y2[ft.fftblockpad * MAXDIM * omp_get_max_threads()];

	// ---------------------------------------------------------------------
	//  note: args x1, x2 must be different arrays
	//  note: args for cfftsx are (direction, layout, xin, xout, scratch)
	//        xin/xout may be the same and it can be somewhat faster
	//        if they are
	// ---------------------------------------------------------------------

	if (dir == 1) {
		cffts1(ft, 1, x1, x1, y1, y2);
		cffts2(ft, 1, x1, x1, y1, y2);
		cffts3(ft, 1, x1, x2, y1, y2);
	} else {
		cffts3(ft, -1, x1, x1, y1, y2);
		cffts2(ft, -1, x1, x1, y1, y2);
		cffts1(ft, -1, x1, x2, y1, y2);
	}
}

static void checksum(const struct FT ft, const int i)
{
	double complex chk = 0;

	// clang-format off
	#pragma omp parallel for reduction(+ : chk)
	// clang-format on
	for (int j = 1; j <= 1024; j++) {
		const int q = j % ft.d1;
		const int r = (3 * j) % ft.d2;
		const int s = (5 * j) % ft.d3;

		// clang-format off
		#define index ixFT(ft, q, r, s)
		// clang-format on

		chk += ft.u1[index];

		// clang-format off
		#undef index
		// clang-format on
	}

	chk /= NTOTAL_F;

	printf(" T = %d     Checksum = %.10E %.10E\n", i, __real__ chk, __imag__ chk);
}

int main(void)
{
	struct FT ft = setup();
	init_ui(ft);

	// ---------------------------------------------------------------------
	// Run the entire problem once to make sure all data is touched.
	// This reduces variable startup costs, which is important for such a
	// short benchmark. The other NPB 2 implementations are similar.
	// ---------------------------------------------------------------------
	compute_indexmap(ft);
	compute_initial_conditions(ft);
	fft_init(ft);
	fft(ft, 1, ft.u1, ft.u0);

	// ---------------------------------------------------------------------
	// Start over from the beginning. Note that all operations must
	// be timed, in contrast to other benchmarks.
	// ---------------------------------------------------------------------
	compute_indexmap(ft);
	compute_initial_conditions(ft);
	fft_init(ft);
	fft(ft, 1, ft.u1, ft.u0);

	for (int iter = 1; iter <= ft.niter; iter++) {
		evolve(ft);
		fft(ft, -1, ft.u1, ft.u1);
		checksum(ft, iter);
	}

	free(ft.u);
	free(ft.u0);
	free(ft.u1);
	free(ft.twiddle);

	return EXIT_SUCCESS;
}
